import random

import numpy as np
import torch


def set_random_seed(seed):
    """Set random seed."""

    seed = int(seed)

    # Set random seed for random
    random.seed(seed)
    # Set random seed for all numpy operations
    np.random.seed(seed=seed)

    # Set random seed for PyTorch
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)


def list_to_tensor(item_list):
    """Convert the list of torch.Tensor into a torch.Tensor."""

    # Convert single value tensor
    if all(item_list[i].dim() == 0 for i in range(len(item_list))):
        item_tensor = torch.stack(item_list, dim=0)
    # Convert 2 or more-D tensor with the same shape
    elif all(
        (item_list[i].size() == item_list[0].size()) and (len(item_list[i].size()) != 1)
        for i in range(len(item_list))
    ):
        item_tensor = torch.stack(item_list, dim=0)
    # Convert reshape to 1-D tensor and then convert
    else:
        item_tensor, _ = pad_batch([item.view(-1) for item in item_list])

    return item_tensor


def pad_batch(batch, max_len=0, pad_value=0, left_padded=False):
    """Convert the batch into a padded tensor and mask tensor.

    :param batch: The data for padding.
    :type batch: list of torch.Tensor
    :param max_len: Max length of sequence of padding.
    :type max_len: int
    :param pad_value: The value to use for padding
    :type pad_value: int
    :param left_padding: if True, pad on the left, otherwise on the right.
    :type left_padding: boolean
    :return: The padded matrix and correspoing mask matrix.
    :rtype: pair of torch.Tensors with shape (batch_size, max_seq_len)
    """

    batch_size = len(batch)
    max_seq_len = int(np.max([len(item) for item in batch]))

    if max_len > 0 and max_len < max_seq_len:
        max_seq_len = max_len

    padded_batch = batch[0].new_full((batch_size, max_seq_len), pad_value)

    for i, item in enumerate(batch):
        length = min(len(item), max_seq_len)
        if left_padded:
            padded_batch[i, -length:] = item[-length:]
        else:
            padded_batch[i, :length] = item[:length]

    mask_batch = torch.eq(padded_batch.clone().detach(), pad_value).type_as(
        padded_batch
    )

    return padded_batch, mask_batch


def prob_to_pred(probs):
    """Identify the class with the maximum probability (add 1 since we assume label
    class starts from 1)

    :param probs: probabilities
    :type probs: np.array
    """

    return np.argmax(probs, axis=-1) + 1


def pred_to_prob(preds, n_classes):
    """Converts predicted labels to probabilistic labels

    :param preds: predicted labels
    :type probs: np.array
    """

    preds = preds.reshape(-1)
    probs = np.zeros((preds.shape[0], n_classes))

    for idx, class_idx in enumerate(preds):
        probs[idx, class_idx - 1] = 1.0

    return probs


def move_to_device(obj, device=-1):
    """
    Given a structure (possibly) containing Tensors on the CPU, move all the Tensors
    to the specified GPU (or do nothing, if they should beon the CPU).
    device = -1 -> "cpu"
    device =  0 -> "cuda:0"
    Originally from:
    https://github.com/HazyResearch/metal/blob/mmtl_clean/metal/utils.py
    """

    if device < 0 or not torch.cuda.is_available():
        return obj
    elif isinstance(obj, torch.Tensor):
        return obj.cuda(device)
    elif isinstance(obj, dict):
        return {key: move_to_device(value, device) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [move_to_device(item, device) for item in obj]
    elif isinstance(obj, tuple):
        return tuple([move_to_device(item, device) for item in obj])
    else:
        return obj


def merge(x, y):
    """Merge two nested dictionaries. Overwrite values in x with values in y."""

    merged = {**x, **y}

    xkeys = x.keys()

    for key in xkeys:
        if isinstance(x[key], dict) and key in y:
            merged[key] = merge(x[key], y[key])

    return merged


def str2bool(v):
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise ValueError("Boolean value expected.")


def str2dict(v):
    dict = {}
    for token in v.split(","):
        key, value = token.split(":")
        dict[key] = value

    return dict



import logging

from emmental.metrics import METRICS

logger = logging.getLogger(__name__)


class Scorer(object):
    """A class to score tasks

    :param metrics: a list of metric names which provides in emmental (e.g., accuracy)
    :type metrics: list
    :param customize_metric_funcs: a dict of customize metric where key is the metric
        name and value is the metric function which takes gold, preds, probs as input
    :type customize_metric_funcs: dict
    """

    def __init__(self, metrics=[], customize_metric_funcs={}):
        self.metrics = dict()
        for metric in metrics:
            if metric not in METRICS:
                raise ValueError(f"Unrecognized metric: {metric}")
            self.metrics[metric] = METRICS[metric]

        self.metrics.update(customize_metric_funcs)

    def score(self, gold, preds, probs, uids=None):
        metric_dict = dict()

        for metric_name, metric in self.metrics.items():
            # handle no examples
            if len(gold) == 0:
                metric_dict[metric_name] = float("nan")
                continue

            res = metric(gold, preds, probs, uids)

            if isinstance(res, dict):
                metric_dict.update(res)
            else:
                metric_dict[metric_name] = res

        return metric_dict



from sklearn.metrics import matthews_corrcoef


def matthews_correlation_coefficient_scorer(golds, probs, preds, uids=None):
    """Matthews correlation coefficient (MCC).

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities. (Not used!)
    :type probs: k-d np.array
    :param preds: Predicted target values.
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :return: Matthews correlation coefficient score.
    :rtype: dict
    """

    return {"matthews_corrcoef": matthews_corrcoef(golds, preds)}



import json
import os
from collections import defaultdict

import yaml

from emmental.meta import Meta


class LogWriter(object):
    """A class for logging during training process.
    """

    def __init__(self):
        """Initialize the log writer.
        """

        self.run_log = defaultdict(list)

    def add_config(self, config):
        """Log config.

        :param config: The config
        :type config: dict
        """

        self.config = config

    def add_scalar(self, name, value, step):
        """Log a scalar variable.

        :param name: The name of the scalar
        :type name: str
        :param value: The value of the scalar
        :type value: float or int
        :param step: The current step
        :type step: float or int
        """

        self.run_log[name].append((step, value))

    def write_config(self, config_filename="config.yaml"):
        """Dump the config to file

        :param config_filename: The config filename, defaults to "config.yaml"
        :type config_filename: str, optional
        """

        config_path = os.path.join(Meta.log_path, config_filename)
        with open(config_path, "w") as yml:
            yaml.dump(self.config, yml, default_flow_style=False, allow_unicode=True)

    def write_log(self, log_filename="log.json"):
        """Dump the log to file.

        :param log_filename: The log filename, defaults to "log.json"
        :type log_filename: str, optional
        """

        log_path = os.path.join(Meta.log_path, log_filename)
        print(log_path)
        with open(log_path, "w") as f:
            json.dump(self.run_log, f)

    def close(self):
        """Close the log writer.
        """
        pass



import numpy as np

from emmental.utils.utils import prob_to_pred


def precision_scorer(golds, probs, preds, uids=None, pos_label=1):
    """Precision.

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities. (Not used!)
    :type probs: k-d np.array
    :param preds: Predicted target values.
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :return: Precision.
    :rtype: dict
    """
    if len(golds.shape) > 1:
        golds = prob_to_pred(golds)
    pred_pos = np.where(preds == pos_label, True, False)
    gt_pos = np.where(golds == pos_label, True, False)
    TP = np.sum(pred_pos * gt_pos)
    FP = np.sum(pred_pos * np.logical_not(gt_pos))

    precision = TP / (TP + FP) if TP + FP > 0 else 0.0

    return {"precision": precision}



from emmental.schedulers.scheduler import Scheduler


class SequentialScheduler(Scheduler):
    """Generate batch generator from all dataloaders in sequential order for MTL
    training.
    """

    def __init__(self):
        super().__init__()

    def get_batches(self, dataloaders):
        """Generate batch generator from all dataloaders in sequential order for
        one epoch.

        :param dataloaders: a list of dataloaders
        :type dataloaders: list
        :return: A generator of all batches
        :rtype: genertor
        """

        task_to_label_dicts = [
            dataloader.task_to_label_dict for dataloader in dataloaders
        ]
        data_names = [dataloader.data_name for dataloader in dataloaders]
        batch_counts = [len(dataloader) for dataloader in dataloaders]
        data_loaders = [iter(dataloader) for dataloader in dataloaders]
        splits = [dataloader.split for dataloader in dataloaders]

        for task_to_label_dict, data_name, batch_count, data_loader, split in zip(
            task_to_label_dicts, data_names, batch_counts, data_loaders, splits
        ):
            for batch in range(batch_count):
                yield next(data_loader), task_to_label_dict, data_name, split



import logging
from collections import defaultdict

import torch
from torch.utils.data import DataLoader, Dataset

from emmental.utils.utils import list_to_tensor

logger = logging.getLogger(__name__)


class EmmentalDataset(Dataset):
    """An advanced dataset class to handle that the input data contains mulitple
    fields and the output data contains multiple label sets

    :param name: the name of the dataset
    :type name: str
    :param X_dict: the feature dict where key is the feature name and value is the
        feature
    :type X_dict: dict
    :param Y_dict: the label dict where key is the label name and value is
        the label
    :type Y_dict: dict
    :param uid: the unique id key in the X_dict (default: None)
    :type uid: str
    """

    def __init__(self, name, X_dict, Y_dict, uid=None):
        self.name = name
        self.X_dict = X_dict
        self.Y_dict = Y_dict
        self.uid = uid

        if self.uid and self.uid not in self.X_dict:
            raise ValueError(f"Cannot find {self.uid} in X_dict.")

        for name, label in self.Y_dict.items():
            if not isinstance(label, torch.Tensor):
                raise ValueError(
                    f"Label {name} should be torch.Tensor, not {type(label)}."
                )

    def __getitem__(self, index):
        x_dict = {name: feature[index] for name, feature in self.X_dict.items()}
        y_dict = {name: label[index] for name, label in self.Y_dict.items()}
        return x_dict, y_dict

    def __len__(self):
        try:
            return len(next(iter(self.X_dict.values())))
        except StopIteration:
            return 0

    def _update_dict(self, ori_dict, new_dict):
        for key, value in new_dict.items():
            ori_dict[key] = value

    def _remove_key(self, ori_dict, key):
        if key in ori_dict:
            del ori_dict[key]

    def add_features(self, X_dict):
        """Add new features into X_dict

        :param X_dict: the new feature dict to add into the existing feature dict
        :type X_dict: dict
        """

        self._update_dict(self.X_dict, X_dict)

    def add_labels(self, Y_dict):
        """Add new labels into Y_dict

        :param Y_dict: the new label dict to add into the existing label dict
        :type Y_dict: dict
        """

        for name, label in Y_dict.items():
            if not isinstance(label, torch.Tensor):
                raise ValueError(f"Label {name} should be torch.Tensor.")

        self._update_dict(self.Y_dict, Y_dict)

    def remove_feature(self, feature_name):
        """Remove one feature from feature dict

        :param feature_name: the feature that removes from feature dict
        :type feature_name: str
        """

        self._remove_key(self.X_dict, feature_name)

    def remove_label(self, label_name):
        """Remove one label from label dict

        :param label_name: the label that removes from label dict
        :type label_name: str
        """

        self._remove_key(self.Y_dict, label_name)


def emmental_collate_fn(batch):

    X_batch = defaultdict(list)
    Y_batch = defaultdict(list)

    for x_dict, y_dict in batch:
        for field_name, value in x_dict.items():
            X_batch[field_name].append(value)
        for label_name, value in y_dict.items():
            Y_batch[label_name].append(value)

    for field_name, values in X_batch.items():
        # Only merge list of tensors
        if isinstance(values[0], torch.Tensor):
            X_batch[field_name] = list_to_tensor(values)

    for label_name, values in Y_batch.items():
        Y_batch[label_name] = list_to_tensor(values)

    return dict(X_batch), dict(Y_batch)


class EmmentalDataLoader(DataLoader):
    """An advanced dataloader class which contains mapping from task to label (which
    label(s) to use in dataset's Y_dict for this task), and split (which part this
    dataset belongs to) information.

    :param task_to_label_dict: the task to label mapping where key is the task name
        and value is the label(s) for that task and should be the key in Y_dict
    :type task_to_label_dict: dict
    :param dataset: the dataset to construct the dataloader
    :type dataset: EmmentalDataset
    :param split: the split information, defaults to "train"
    :type split: str, optional
    :param collate_fn: the function that merges a list of samples to form a
        mini-batch, defaults to emmental_collate_fn
    :type collate_fn: function, optional
    """

    def __init__(
        self,
        task_to_label_dict,
        dataset,
        split="train",
        collate_fn=emmental_collate_fn,
        **kwargs,
    ):

        assert isinstance(dataset, EmmentalDataset)
        super().__init__(dataset, collate_fn=collate_fn, **kwargs)

        self.task_to_label_dict = task_to_label_dict
        self.data_name = dataset.name
        self.split = split

        for task_name, label_names in task_to_label_dict.items():
            if not isinstance(label_names, list):
                label_names = [label_names]
            unrecognized_labels = set(label_names) - set(dataset.Y_dict.keys())
            if len(unrecognized_labels) > 0:
                msg = (
                    f"Unrecognized Label {unrecognized_labels} of Task {task_name} in "
                    f"dataset {dataset.name}"
                )
                logger.warn(msg)
                raise ValueError(msg)



from sklearn.metrics import mean_squared_error


def mean_squared_error_scorer(golds, probs, preds, uids=None):
    """Mean squared error regression loss.

    :param golds: Ground truth (correct) target values.
    :type golds: k-d np.array
    :param probs: Predicted target probabilities.
    :type probs: k-d np.array
    :param preds: Predicted target values. (Not used!)
    :type preds: any
    :param uids: Unique ids.
    :type uids: list
    :return: Mean squared error regression loss.
    :rtype: dict
    """

    return {"mean_squared_error": float(mean_squared_error(golds, probs))}



import logging
from collections import defaultdict

import torch
import torch.optim as optim

from emmental import Meta
from emmental.logging import LoggingManager
from emmental.optimizers.bert_adam import BertAdam
from emmental.schedulers.round_robin_scheduler import RoundRobinScheduler
from emmental.schedulers.sequential_scheduler import SequentialScheduler

try:
    from IPython import get_ipython

    if "IPKernelApp" not in get_ipython().config:
        raise ImportError("console")
except (AttributeError, ImportError):
    from tqdm import tqdm
else:
    from tqdm import tqdm_notebook as tqdm

logger = logging.getLogger(__name__)


class EmmentalLearner(object):
    """A class for emmental multi-task learning.
    """

    def __init__(self, name=None):
        self.name = name if name is not None else type(self).__name__

    def _set_logging_manager(self):
        """Set logging manager."""

        self.logging_manager = LoggingManager(self.n_batches_per_epoch)

    def _set_optimizer(self, model):
        """Set optimizer for learning process."""

        # TODO: add more optimizer support and fp16
        optimizer_config = Meta.config["learner_config"]["optimizer_config"]
        opt = optimizer_config["optimizer"]

        parameters = filter(lambda p: p.requires_grad, model.parameters())

        if opt == "sgd":
            optimizer = optim.SGD(
                parameters,
                lr=optimizer_config["lr"],
                **optimizer_config["sgd_config"],
                weight_decay=optimizer_config["l2"],
            )
        elif opt == "adam":
            optimizer = optim.Adam(
                parameters,
                lr=optimizer_config["lr"],
                **optimizer_config["adam_config"],
                weight_decay=optimizer_config["l2"],
            )
        elif opt == "adamax":
            optimizer = optim.Adamax(
                parameters,
                lr=optimizer_config["lr"],
                **optimizer_config["adamax_config"],
                weight_decay=optimizer_config["l2"],
            )
        elif opt == "bert_adam":
            optimizer = BertAdam(
                parameters,
                lr=optimizer_config["lr"],
                **optimizer_config["bert_adam_config"],
                weight_decay=optimizer_config["l2"],
            )
        else:
            raise ValueError(f"Unrecognized optimizer option '{opt}'")

        logger.info(f"Using optimizer {optimizer}")

        self.optimizer = optimizer

    def _set_lr_scheduler(self, model):
        """Set learning rate scheduler for learning process."""

        # Set warmup scheduler
        self._set_warmup_scheduler(model)

        # Set lr scheduler
        # TODO: add more lr scheduler support
        opt = Meta.config["learner_config"]["lr_scheduler_config"]["lr_scheduler"]
        lr_scheduler_config = Meta.config["learner_config"]["lr_scheduler_config"]

        if opt is None:
            lr_scheduler = None
        elif opt == "linear":
            total_steps = (
                self.n_batches_per_epoch * Meta.config["learner_config"]["n_epochs"]
            )
            linear_decay_func = lambda x: (total_steps - self.warmup_steps - x) / (
                total_steps - self.warmup_steps
            )
            lr_scheduler = optim.lr_scheduler.LambdaLR(
                self.optimizer, linear_decay_func
            )
        elif opt == "exponential":
            lr_scheduler = optim.lr_scheduler.ExponentialLR(
                self.optimizer, **lr_scheduler_config["exponential_config"]
            )
        elif opt == "step":
            lr_scheduler = optim.lr_scheduler.StepLR(
                self.optimizer, **lr_scheduler_config["step_config"]
            )
        elif opt == "multi_step":
            lr_scheduler = optim.lr_scheduler.MultiStepLR(
                self.optimizer, **lr_scheduler_config["multi_step_config"]
            )
        # elif opt == "reduce_on_plateau":
        #     lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(
        #         self.optimizer,
        #         min_lr=lr_scheduler_config["min_lr"],
        #         **lr_scheduler_config["plateau_config"],
        #     )
        else:
            raise ValueError(f"Unrecognized lr scheduler option '{opt}'")

        self.lr_scheduler = lr_scheduler

    def _set_warmup_scheduler(self, model):
        """Set warmup learning rate scheduler for learning process."""

        if Meta.config["learner_config"]["lr_scheduler_config"]["warmup_steps"]:
            warmup_steps = Meta.config["learner_config"]["lr_scheduler_config"][
                "warmup_steps"
            ]
            if warmup_steps < 0:
                raise ValueError(f"warmup_steps much greater or equal than 0.")
            warmup_unit = Meta.config["learner_config"]["lr_scheduler_config"][
                "warmup_unit"
            ]
            if warmup_unit == "epoch":
                self.warmup_steps = int(warmup_steps * self.n_batches_per_epoch)
            elif warmup_unit == "batch":
                self.warmup_steps = int(warmup_steps)
            else:
                raise ValueError(
                    f"warmup_unit must be 'batch' or 'epoch', but {warmup_unit} found."
                )
            linear_warmup_func = lambda x: x / self.warmup_steps
            warmup_scheduler = optim.lr_scheduler.LambdaLR(
                self.optimizer, linear_warmup_func
            )
            logger.info(f"Warmup {self.warmup_steps} batchs.")
        elif Meta.config["learner_config"]["lr_scheduler_config"]["warmup_percentage"]:
            warmup_percentage = Meta.config["learner_config"]["lr_scheduler_config"][
                "warmup_percentage"
            ]
            self.warmup_steps = int(
                warmup_percentage
                * Meta.config["learner_config"]["n_epochs"]
                * self.n_batches_per_epoch
            )
            linear_warmup_func = lambda x: x / self.warmup_steps
            warmup_scheduler = optim.lr_scheduler.LambdaLR(
                self.optimizer, linear_warmup_func
            )
            logger.info(f"Warmup {self.warmup_steps} batchs.")
        else:
            warmup_scheduler = None
            self.warmup_steps = 0

        self.warmup_scheduler = warmup_scheduler

    def _update_lr_scheduler(self, model, step):
        """Update the lr using lr_scheduler with each batch."""

        if self.warmup_scheduler and step < self.warmup_steps:
            self.warmup_scheduler.step()
        elif self.lr_scheduler is not None:
            self.lr_scheduler.step()
            min_lr = Meta.config["learner_config"]["lr_scheduler_config"]["min_lr"]
            if min_lr and self.optimizer.param_groups[0]["lr"] < min_lr:
                self.optimizer.param_groups[0]["lr"] = min_lr

    def _set_task_scheduler(self, model, dataloaders):
        """Set task scheduler for learning process"""
        # TODO: add more task scheduler support
        opt = Meta.config["learner_config"]["task_scheduler"]

        if opt == "sequential":
            self.task_scheduler = SequentialScheduler()
        elif opt == "round_robin":
            self.task_scheduler = RoundRobinScheduler()
        else:
            raise ValueError(f"Unrecognized task scheduler option '{opt}'")

    def _evaluate(self, model, dataloaders, split):
        if not isinstance(split, list):
            valid_split = [split]
        else:
            valid_split = split

        valid_dataloaders = [
            dataloader for dataloader in dataloaders if dataloader.split in valid_split
        ]
        return model.score(valid_dataloaders)

    def _logging(self, model, dataloaders, batch_size):
        """Checking if it's time to evaluting or checkpointing"""

        # Switch to eval mode for evaluation
        model.eval()

        metric_dict = dict()

        self.logging_manager.update(batch_size)

        # Log the loss and lr
        metric_dict.update(self._aggregate_losses())

        # Evaluate the model and log the metric
        if self.logging_manager.trigger_evaluation():

            # Log task specific metric
            metric_dict.update(
                self._evaluate(
                    model, dataloaders, Meta.config["learner_config"]["valid_split"]
                )
            )

            self.logging_manager.write_log(metric_dict)

            self._reset_losses()

        # Checkpoint the model
        if self.logging_manager.trigger_checkpointing():
            self.logging_manager.checkpoint_model(
                model, self.optimizer, self.lr_scheduler, metric_dict
            )

            self.logging_manager.write_log(metric_dict)

            self._reset_losses()

        # Switch to train mode
        model.train()

        return metric_dict

    def _aggregate_losses(self):
        """Calculate the task specific loss, average micro loss and learning rate."""

        metric_dict = dict()

        # Log task specific loss
        for identifier in self.running_losses.keys():
            if self.running_counts[identifier] > 0:
                metric_dict[identifier] = (
                    self.running_losses[identifier] / self.running_counts[identifier]
                )

        # Calculate average micro loss
        total_loss = sum(self.running_losses.values())
        total_count = sum(self.running_counts.values())
        if total_count > 0:
            metric_dict["model/train/all/loss"] = total_loss / total_count

        # Log the learning rate
        metric_dict["model/train/all/lr"] = self.optimizer.param_groups[0]["lr"]

        return metric_dict

    def _reset_losses(self):
        self.running_losses = defaultdict(float)
        self.running_counts = defaultdict(int)

    def learn(self, model, dataloaders):
        """The learning procedure of emmental MTL

        :param model: The emmental model that needs to learn
        :type model: emmental.model
        :param dataloaders: a list of dataloaders used to learn the model
        :type dataloaders: list
        """

        # Generate the list of dataloaders for learning process
        train_split = Meta.config["learner_config"]["train_split"]
        if isinstance(train_split, str):
            train_split = [train_split]

        train_dataloaders = [
            dataloader for dataloader in dataloaders if dataloader.split in train_split
        ]

        if not train_dataloaders:
            raise ValueError(
                f"Cannot find the specified train_split "
                f'{Meta.config["learner_config"]["train_split"]} in dataloaders.'
            )

        # Calculate the total number of batches per epoch
        self.n_batches_per_epoch = sum(
            [len(dataloader) for dataloader in train_dataloaders]
        )

        # Set up logging manager
        self._set_logging_manager()
        # Set up optimizer
        self._set_optimizer(model)
        # Set up lr_scheduler
        self._set_lr_scheduler(model)
        # Set up task_scheduler
        self._set_task_scheduler(model, dataloaders)

        # Set to training mode
        model.train()

        logger.info(f"Start learning...")

        self.metrics = dict()
        self._reset_losses()

        for epoch_num in range(Meta.config["learner_config"]["n_epochs"]):
            batches = tqdm(
                enumerate(self.task_scheduler.get_batches(train_dataloaders)),
                total=self.n_batches_per_epoch,
                disable=(not Meta.config["meta_config"]["verbose"]),
                desc=f"Epoch {epoch_num}:",
            )
            for batch_num, (batch, task_to_label_dict, data_name, split) in batches:
                X_dict, Y_dict = batch

                total_batch_num = epoch_num * self.n_batches_per_epoch + batch_num
                batch_size = len(next(iter(Y_dict.values())))

                # Update lr using lr scheduler
                self._update_lr_scheduler(model, total_batch_num)

                # Set gradients of all model parameters to zero
                self.optimizer.zero_grad()

                # Perform forward pass and calcualte the loss and count
                loss_dict, count_dict = model.calculate_loss(
                    X_dict, Y_dict, task_to_label_dict, data_name, split
                )

                # Update running loss and count
                for identifier in loss_dict.keys():
                    self.running_losses[identifier] += (
                        loss_dict[identifier].item() * count_dict[identifier]
                    )
                    self.running_counts[identifier] += count_dict[identifier]

                # Skip the backward pass if no loss is calcuated
                if not loss_dict:
                    continue

                # Calculate the average loss
                loss = sum(loss_dict.values())

                # Perform backward pass to calculate gradients
                loss.backward()

                # Clip gradient norm
                if Meta.config["learner_config"]["optimizer_config"]["grad_clip"]:
                    torch.nn.utils.clip_grad_norm_(
                        model.parameters(),
                        Meta.config["learner_config"]["optimizer_config"]["grad_clip"],
                    )

                # Update the parameters
                self.optimizer.step()

                self.metrics.update(self._logging(model, dataloaders, batch_size))

                batches.set_postfix(self.metrics)

        model = self.logging_manager.close(model)



from abc import ABC, abstractmethod


class Scheduler(ABC):
    """Generate batch generator from all dataloaders in designed order for MTL
    training.
    """

    def __init__(self):

        pass

    @abstractmethod
    def get_batches(self, dataloaders):
        """Generate batch generator from all dataloaders in designed order for
        one epoch.

        :param dataloaders: a list of dataloaders
        :type dataloaders: list
        """

        pass



from sklearn.metrics import roc_auc_score

from emmental.utils.utils import pred_to_prob


def roc_auc_scorer(golds, probs, preds, uids=None, pos_label=1):
    """ROC AUC.

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities.
    :type probs: k-d np.array
    :param preds: Predicted target values. (Not used!)
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :return: Recall.
    :rtype: dict
    """

    gold_probs = pred_to_prob(golds, n_classes=probs.shape[1])
    roc_auc = roc_auc_score(gold_probs, probs)

    return {"roc_auc": roc_auc}



import numpy as np
from scipy.stats import pearsonr


def pearson_correlation_scorer(golds, probs, preds, uids=None):
    """Pearson correlation coefficient and the p-value.

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities.
    :type probs: 1-d np.array
    :param preds: Predicted target values. (Not used!)
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :return: Pearson correlation coefficient and the p-value.
    :rtype: dict
    """

    probs = np.vstack(probs).squeeze()
    correlation, pvalue = pearsonr(golds, probs)
    if np.isnan(correlation):
        correlation = 0.0
        pvalue = 0.0
    return {"pearson_correlation": correlation, "pearson_pvalue": pvalue}



import logging
import os
import tempfile
from builtins import object
from datetime import datetime

import yaml

from emmental.utils.utils import merge, set_random_seed

MAX_CONFIG_SEARCH_DEPTH = 25  # Max num of parent directories to look for config

logger = logging.getLogger(__name__)


def init(
    log_dir=tempfile.gettempdir(),
    log_name="emmental.log",
    format="[%(asctime)s][%(levelname)s] %(name)s:%(lineno)s - %(message)s",
    level=logging.INFO,
    config={},
    config_dir=None,
    config_name="emmental-config.yaml",
):
    """Initialize the logging and configuration.
    :param log_dir: The directory to store logs in.
    :type log_dir: str
    :param format: The logging format string to use.
    :type format: str
    :param level: The logging level to use, e.g., logging.INFO.
    :param config: The new configuration, defaults to {}
    :type config: dict, optional
    :param config_dir: the path to the config file, defaults to None
    :type config_dir: str, optional
    :param config_name: the config file name, defaults to "emmental-config.yaml"
    :type config_name: str, optional
    """

    init_logging(log_dir, log_name, format, level)
    init_config()
    if config or config_dir is not None:
        Meta.update_config(config, config_dir, config_name)

    set_random_seed(Meta.config["meta_config"]["seed"])


def init_config():
    """Load the default configuration."""

    # Load the default setting
    default_config_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "emmental-default-config.yaml"
    )
    with open(default_config_path, "r") as f:
        config = yaml.load(f)
    logger.info(f"Loading Emmental default config from {default_config_path}.")

    Meta.config = config


def init_logging(
    log_dir=tempfile.gettempdir(),
    log_name="emmental.log",
    format="[%(asctime)s][%(levelname)s] %(name)s:%(lineno)s - %(message)s",
    level=logging.INFO,
):
    """Configures logging to output to the provided log_dir.
    Will use a nested directory whose name is the current timestamp.
    :param log_dir: The directory to store logs in.
    :type log_dir: str
    :param format: The logging format string to use.
    :type format: str
    :param level: The logging level to use, e.g., logging.INFO.
    """

    if not Meta.log_path:
        # Generate a new directory using the log_dir, if it doesn't exist
        date = datetime.now().strftime("%Y_%m_%d")
        time = datetime.now().strftime("%H_%M_%S")
        log_path = os.path.join(log_dir, date, time)
        if not os.path.exists(log_path):
            os.makedirs(log_path)

        # Configure the logger using the provided path
        logging.basicConfig(
            format=format,
            level=level,
            handlers=[
                logging.FileHandler(os.path.join(log_path, log_name)),
                logging.StreamHandler(),
            ],
        )

        # Notify user of log location
        logger.info(f"Setting logging directory to: {log_path}")
        Meta.log_path = log_path
    else:
        logger.info(
            f"Logging was already initialized to use {Meta.log_path}.  "
            "To configure logging manually, call emmental.init_logging before "
            "initialiting Meta."
        )


class Meta(object):
    """Singleton-like metadata class for all global variables.
    Adapted from the Unique Design Pattern:
    https://stackoverflow.com/questions/1318406/why-is-the-borg-pattern-better-than-the-singleton-pattern-in-python
    """

    log_path = None
    config = None

    @classmethod
    def init(cls):
        """Return the unique Meta class."""
        if not Meta.log_path:
            init_logging()

        if not Meta.config:
            init_config()

        return cls

    def update_config(config={}, path=None, filename="emmental-config.yaml"):
        """Update the configuration with the configs in root of project and
        its parents.

        Note: There are two ways to update the config:
            (1) uses a config dict to update to config
            (2) uses path and filename to load yaml file to update config

        :param config: The new configuration, defaults to {}
        :type config: dict, optional
        :param path: the path to the config file, defaults to os.getcwd()
        :param path: str, optional
        :param filename: the config file name, defaults to "emmental-config.yaml"
        :param filename: str, optional
        """

        if config != {}:
            Meta.config = merge(Meta.config, config)
            logger.info("Updating Emmental config from user provided config.")

        if path is not None:
            tries = 0
            current_dir = path
            while current_dir and tries < MAX_CONFIG_SEARCH_DEPTH:
                potential_path = os.path.join(current_dir, filename)
                if os.path.exists(potential_path):
                    with open(potential_path, "r") as f:
                        Meta.config = merge(Meta.config, yaml.safe_load(f))
                    logger.info(f"Updating Emmental config from {potential_path}.")
                    break

                new_dir = os.path.split(current_dir)[0]
                if current_dir == new_dir:
                    logger.info("Unable to find config file. Using defaults.")
                    break
                current_dir = new_dir
                tries += 1

    def reset():
        """ Clears shared variables of shared, global singleton. """

        Meta.log_path = None
        Meta.config = None



import random

from emmental.schedulers.scheduler import Scheduler


class RoundRobinScheduler(Scheduler):
    """Generate batch generator from all dataloaders in round robin order for MTL
    training.
    """

    def __init__(self):
        super().__init__()

    def get_batches(self, dataloaders):
        """Generate batch generator from all dataloaders in round robin order for
        one epoch.

        :param dataloaders: a list of dataloaders
        :type dataloaders: list
        :return: A generator of all batches
        :rtype: genertor
        """

        task_to_label_dicts = [
            dataloader.task_to_label_dict for dataloader in dataloaders
        ]
        data_names = [dataloader.data_name for dataloader in dataloaders]
        batch_counts = [len(dataloader) for dataloader in dataloaders]
        data_loaders = [iter(dataloader) for dataloader in dataloaders]
        splits = [dataloader.split for dataloader in dataloaders]

        dataloader_indexer = []
        for idx, count in enumerate(batch_counts):
            dataloader_indexer.extend([idx] * count)

        random.shuffle(dataloader_indexer)

        for index in dataloader_indexer:
            yield next(data_loaders[index]), task_to_label_dicts[index], data_names[
                index
            ], splits[index]



from emmental.metrics.precision import precision_scorer
from emmental.metrics.recall import recall_scorer


def fbeta_scorer(golds, probs, preds, uids=None, pos_label=1, beta=1):
    """F-beta score is the weighted harmonic mean of precision and recall.

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities. (Not used!)
    :type probs: k-d np.array
    :param preds: Predicted target values.
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :param pos_label: The positive class label, defaults to 1
    :param pos_label: int, optional
    :param beta: Weight of precision in harmonic mean, defaults to 1
    :param beta: float, optional
    :return: F-beta score.
    :rtype: dict
    """

    precision = precision_scorer(golds, probs, preds, uids, pos_label)["precision"]
    recall = recall_scorer(golds, probs, preds, uids, pos_label)["recall"]

    fbeta = (
        (1 + beta ** 2) * (precision * recall) / ((beta ** 2 * precision) + recall)
        if (beta ** 2 * precision) + recall > 0
        else 0.0
    )

    return {f"f{beta}": fbeta}


def f1_scorer(golds, probs, preds, uids=None, pos_label=1):
    return {"f1": fbeta_scorer(golds, probs, preds, uids, pos_label, beta=1)["f1"]}



import logging
import os
from collections import defaultdict
from collections.abc import Iterable

import numpy as np
import torch
import torch.nn as nn

from emmental.meta import Meta
from emmental.task import EmmentalTask
from emmental.utils.utils import move_to_device, prob_to_pred

logger = logging.getLogger(__name__)


class EmmentalModel(nn.Module):
    """A class to build multi-task model.

    :param name: Name of the model
    :type name: str
    :param tasks: a list of Tasks that trains jointly
    :type tasks: list of Task project
    """

    def __init__(self, name=None, tasks=None):
        super().__init__()
        self.name = name if name is not None else type(self).__name__

        # Initiate the model attributes
        self.module_pool = nn.ModuleDict()
        self.task_names = set()
        self.task_flows = dict()
        self.loss_funcs = dict()
        self.output_funcs = dict()
        self.scorers = dict()

        # Build network with given tasks
        if tasks is not None:
            self._build_network(tasks)

        if Meta.config["meta_config"]["verbose"]:
            logger.info(
                f"Created emmental model {self.name} that contains "
                f"task {self.task_names}."
            )

        # Move model to specified device
        self._move_to_device()

    def _move_to_device(self):
        """Move model to specified device."""

        if Meta.config["model_config"]["device"] >= 0:
            if torch.cuda.is_available():
                if Meta.config["meta_config"]["verbose"]:
                    logger.info(
                        f"Moving model to GPU "
                        f"(cuda:{Meta.config['model_config']['device']})."
                    )
                self.to(torch.device(f"cuda:{Meta.config['model_config']['device']}"))
            else:
                if Meta.config["meta_config"]["verbose"]:
                    logger.info("No cuda device available. Switch to cpu instead.")

    def _build_network(self, tasks):
        """Build the MTL network using all tasks"""

        if not isinstance(tasks, Iterable):
            tasks = [tasks]
        for task in tasks:
            if task.name in self.task_names:
                raise ValueError(
                    f"Found duplicate task {task.name}, different task should use "
                    f"different task name."
                )
            if not isinstance(task, EmmentalTask):
                raise ValueError(f"Unrecognized task type {task}.")
            self.add_task(task)

    def add_task(self, task):
        """Add a single task into MTL network"""

        # Combine module_pool from all tasks
        for key in task.module_pool.keys():
            if key in self.module_pool.keys():
                if Meta.config["model_config"]["dataparallel"]:
                    task.module_pool[key] = nn.DataParallel(self.module_pool[key])
                else:
                    task.module_pool[key] = self.module_pool[key]
            else:
                if Meta.config["model_config"]["dataparallel"]:
                    self.module_pool[key] = nn.DataParallel(task.module_pool[key])
                else:
                    self.module_pool[key] = task.module_pool[key]
        # Collect task names
        self.task_names.add(task.name)
        # Collect task flows
        self.task_flows[task.name] = task.task_flow
        # Collect loss functions
        self.loss_funcs[task.name] = task.loss_func
        # Collect output functions
        self.output_funcs[task.name] = task.output_func
        # Collect scorers
        self.scorers[task.name] = task.scorer

        # Move model to specified device
        self._move_to_device()

    def update_task(self, task):
        """Update a existing task in MTL network"""

        # Update module_pool with task
        for key in task.module_pool.keys():
            # Update the model's module with the task's module
            if Meta.config["model_config"]["dataparallel"]:
                self.module_pool[key] = nn.DataParallel(task.module_pool[key])
            else:
                self.module_pool[key] = task.module_pool[key]
        # Update task flows
        self.task_flows[task.name] = task.task_flow
        # Update loss functions
        self.loss_funcs[task.name] = task.loss_func
        # Update output functions
        self.output_funcs[task.name] = task.output_func
        # Collect scorers
        self.scorers[task.name] = task.scorer

        # Move model to specified device
        self._move_to_device()

    def remove_task(self, task_name):
        """Remove a existing task from MTL network"""
        if task_name not in self.task_flows:
            if Meta.config["meta_config"]["verbose"]:
                logger.info(f"Task ({task_name}) not in the current model, skip...")
            return

        # Remove task by task_name
        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"Removing Task {task_name}.")

        self.task_names.remove(task_name)
        del self.task_flows[task_name]
        del self.loss_funcs[task_name]
        del self.output_funcs[task_name]
        del self.scorers[task_name]
        # TODO: remove the modules only associate with that task

    def __repr__(self):
        cls_name = type(self).__name__
        return f"{cls_name}(name={self.name})"

    def forward(self, X_dict, task_names):
        """Forward based on input and task
            Note: We assume that all shared modules from all tasks are based on the
            the same input.

        :param X_dict: The input data
        :type X_dict: dict of tensor
        :param task_names: The task names that needs to forward
        :type task_names: list of str
        :return: The output of all forwarded modules
        :rtype: dict
        """

        X_dict = move_to_device(X_dict, Meta.config["model_config"]["device"])

        immediate_ouput_dict = dict()
        immediate_ouput_dict["_input_"] = X_dict

        # Call forward for each task
        for task_name in task_names:
            task_flow = self.task_flows[task_name]

            for action in task_flow:
                if action["name"] not in immediate_ouput_dict:
                    if action["inputs"]:
                        try:
                            input = [
                                immediate_ouput_dict[action_name][output_index]
                                for action_name, output_index in action["inputs"]
                            ]
                        except Exception:
                            raise ValueError(f"Unrecognized action {action}.")
                        output = self.module_pool[action["module"]].forward(*input)
                    else:
                        output = self.module_pool[action["module"]].forward(
                            immediate_ouput_dict
                        )
                    if isinstance(output, tuple):
                        output = list(output)
                    if not isinstance(output, list):
                        output = [output]
                    immediate_ouput_dict[action["name"]] = output

        return immediate_ouput_dict

    def calculate_loss(self, X_dict, Y_dict, task_to_label_dict, data_name, split):
        """Calculate the loss

        :param X_dict: The input data
        :type X_dict: dict of tensors
        :param Y_dict: The output data
        :type Y_dict: dict of tensors
        :param task_to_label_dict: The task to label mapping
        :type task_to_label_dict: dict
        :param data_name: The dataset name
        :type data_name: str
        :param split: The data split
        :type split: str
        :return: The loss and the number of samples in the batch of all tasks
        :rtype: dict, dict
        """

        loss_dict = dict()
        count_dict = dict()

        immediate_ouput_dict = self.forward(X_dict, task_to_label_dict.keys())

        # Calculate loss for each task
        for task_name, label_name in task_to_label_dict.items():
            identifier = "/".join([task_name, data_name, split, "loss"])

            Y = Y_dict[label_name]

            # Select the active samples
            if len(Y.size()) == 1:
                active = Y.detach() != Meta.config["learner_config"]["ignore_index"]
            else:
                active = torch.any(
                    Y.detach() != Meta.config["learner_config"]["ignore_index"], dim=1
                )

            # Only calculate the loss when active example exists
            if active.any():
                count_dict[identifier] = active.sum().item()

                loss_dict[identifier] = self.loss_funcs[task_name](
                    immediate_ouput_dict,
                    move_to_device(
                        Y_dict[label_name], Meta.config["model_config"]["device"]
                    ),
                    move_to_device(active, Meta.config["model_config"]["device"]),
                )

        return loss_dict, count_dict

    @torch.no_grad()
    def _calculate_probs(self, X_dict, task_names):
        """Calculate the probs given the features

        :param X_dict: The input data
        :type X_dict: dict of tensor
        :param task_names: The task names that needs to forward
        :type task_names: list of str
        """

        self.eval()

        prob_dict = dict()

        immediate_ouput_dict = self.forward(X_dict, task_names)

        # Calculate prediction for each task
        for task_name in task_names:
            prob_dict[task_name] = (
                self.output_funcs[task_name](immediate_ouput_dict).cpu().numpy()
            )

        return prob_dict

    @torch.no_grad()
    def predict(self, dataloader, return_preds=False, return_uids=False):

        self.eval()

        uid_key = dataloader.dataset.uid

        # Check uid exists
        if return_uids and uid_key is None:
            return_uids = False
            logger.info("No uid exist, skip it...")

        uid_dict = defaultdict(list)
        gold_dict = defaultdict(list)
        prob_dict = defaultdict(list)

        for batch_num, (X_batch_dict, Y_batch_dict) in enumerate(dataloader):
            prob_batch_dict = self._calculate_probs(
                X_batch_dict, dataloader.task_to_label_dict.keys()
            )
            for task_name in dataloader.task_to_label_dict.keys():
                if return_uids:
                    uid_dict[task_name].extend(X_batch_dict[uid_key])
                prob_dict[task_name].extend(prob_batch_dict[task_name])
                gold_dict[task_name].extend(
                    Y_batch_dict[dataloader.task_to_label_dict[task_name]].cpu().numpy()
                )
        for task_name in gold_dict:
            gold_dict[task_name] = np.array(gold_dict[task_name])
            prob_dict[task_name] = np.array(prob_dict[task_name])
            if len(gold_dict[task_name].shape) == 1:
                active = (
                    gold_dict[task_name]
                    != Meta.config["learner_config"]["ignore_index"]
                ).reshape(-1)
            else:
                active = (
                    np.sum(
                        gold_dict[task_name]
                        == Meta.config["learner_config"]["ignore_index"],
                        axis=1,
                    )
                    > 0
                )

            if 0 in active:
                gold_dict[task_name] = gold_dict[task_name][active]
                prob_dict[task_name] = prob_dict[task_name][active]
                if return_uids:
                    uid_dict[task_name] = [
                        uid_dict[task_name][i]
                        for i, value in enumerate(active)
                        if value
                    ]

        if return_preds:
            pred_dict = defaultdict(list)
            for task_name, prob in prob_dict.items():
                pred_dict[task_name] = prob_to_pred(prob)

        res = {"golds": gold_dict, "probs": prob_dict}

        if return_preds:
            res["preds"] = pred_dict

        if return_uids:
            res["uids"] = uid_dict

        return res

    @torch.no_grad()
    def score(self, dataloaders):
        """Score the data from dataloader with the model

        :param dataloaders: the dataloader that performs scoring
        :type dataloaders: dataloader
        """

        self.eval()

        if not isinstance(dataloaders, list):
            dataloaders = [dataloaders]

        metric_score_dict = dict()

        for dataloader in dataloaders:
            return_uids = True if dataloader.dataset.uid else False
            preds = self.predict(dataloader, return_preds=True, return_uids=return_uids)
            for task_name in preds["golds"].keys():
                metric_score = self.scorers[task_name].score(
                    preds["golds"][task_name],
                    preds["probs"][task_name],
                    preds["preds"][task_name],
                    preds["uids"][task_name] if return_uids else None,
                )
                for metric_name, metric_value in metric_score.items():
                    identifier = "/".join(
                        [task_name, dataloader.data_name, dataloader.split, metric_name]
                    )
                    metric_score_dict[identifier] = metric_value

        # TODO: have a better to handle global evaluation metric
        if Meta.config["learner_config"]["global_evaluation_metric_dict"]:
            global_evaluation_metric_dict = Meta.config["learner_config"][
                "global_evaluation_metric_dict"
            ]
            for metric_name, metric in global_evaluation_metric_dict.items():
                metric_score_dict[metric_name] = metric(metric_score_dict)

        return metric_score_dict

    def save(self, model_path):
        """Save the current model

        :param model_path: Saved model path.
        :type model_path: str
        """

        # Check existence of model saving directory and create if does not exist.
        if not os.path.exists(os.path.dirname(model_path)):
            os.makedirs(os.path.dirname(model_path))

        state_dict = {
            "model": {
                "name": self.name,
                "module_pool": self.collect_state_dict(),
                # "task_names": self.task_names,
                # "task_flows": self.task_flows,
                # "loss_funcs": self.loss_funcs,
                # "output_funcs": self.output_funcs,
                # "scorers": self.scorers,
            }
        }

        try:
            torch.save(state_dict, model_path)
        except BaseException:
            logger.warning("Saving failed... continuing anyway.")

        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"[{self.name}] Model saved in {model_path}")

    def load(self, model_path):
        """Load model state_dict from file and reinitialize the model weights.

        :param model_path: Saved model path.
        :type model_path: str
        """

        if not os.path.exists(model_path):
            logger.error("Loading failed... Model does not exist.")

        try:
            checkpoint = torch.load(model_path, map_location=torch.device("cpu"))
        except BaseException:
            logger.error(f"Loading failed... Cannot load model from {model_path}")
            raise

        self.load_state_dict(checkpoint["model"]["module_pool"])

        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"[{self.name}] Model loaded from {model_path}")

        # Move model to specified device
        self._move_to_device()

    def collect_state_dict(self):
        state_dict = defaultdict(list)

        for module_name, module in self.module_pool.items():
            if Meta.config["model_config"]["dataparallel"]:
                state_dict[module_name] = module.module.state_dict()
            else:
                state_dict[module_name] = module.state_dict()

        return state_dict

    def load_state_dict(self, state_dict):

        for module_name, module_state_dict in state_dict.items():
            if module_name in self.module_pool:
                if Meta.config["model_config"]["dataparallel"]:
                    self.module_pool[module_name].module.load_state_dict(
                        module_state_dict
                    )
                else:
                    self.module_pool[module_name].load_state_dict(module_state_dict)
            else:
                logger.info(f"Missing {module_name} in module_pool, skip it..")



import logging

from torch import nn

logger = logging.getLogger(__name__)


class EmmentalTask(object):
    """Task class to define task in Emmental model.

    :param name: The name of the task (Primary key).
    :type name: str
    :param module_pool: A dict of modules that uses in the task.
    :type module_pool: nn.ModuleDict
    :param task_flow: The task flow among modules to define how the data flows.
    :type task_flow: list
    :param loss_func: The function to calculate the loss.
    :type loss_func: function
    :param output_func: The function to generate the output.
    :type output_func: function
    :param scorer: The class of metrics to evaluate the task.
    :type scorer: Scorer class
    """

    def __init__(self, name, module_pool, task_flow, loss_func, output_func, scorer):
        self.name = name
        assert isinstance(module_pool, nn.ModuleDict) is True
        self.module_pool = module_pool
        self.task_flow = task_flow
        self.loss_func = loss_func
        self.output_func = output_func
        self.scorer = scorer

        logger.info(f"Created task: {self.name}")

    def __repr__(self):
        cls_name = type(self).__name__
        return f"{cls_name}(name={self.name})"



import numpy as np

from emmental.utils.utils import prob_to_pred


def recall_scorer(golds, probs, preds, uids=None, pos_label=1):
    """Recall.

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities. (Not used!)
    :type probs: k-d np.array
    :param preds: Predicted target values.
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :return: Recall.
    :rtype: dict
    """
    if len(golds.shape) > 1:
        golds = prob_to_pred(golds)
    pred_pos = np.where(preds == pos_label, True, False)
    gt_pos = np.where(golds == pos_label, True, False)
    TP = np.sum(pred_pos * gt_pos)
    FN = np.sum(np.logical_not(pred_pos) * gt_pos)

    recall = TP / (TP + FN) if TP + FN > 0 else 0.0

    return {"recall": recall}



import numpy as np

from emmental.metrics.pearson_correlation import pearson_correlation_scorer
from emmental.metrics.spearman_correlation import spearman_correlation_scorer


def pearson_spearman_scorer(golds, probs, preds, uids=None):
    """Average of Pearson correlation coefficient and the p-value and Spearman
    rank-order correlation coefficient.

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities.
    :type probs: 1-d np.array
    :param preds: Predicted target values. (Not used!)
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :return: Pearson correlation coefficient, the p-value and Spearman
        rank-order correlation coefficient and the average.
    :rtype: dict
    """

    metrics = dict()
    metrics.update(pearson_correlation_scorer(golds, probs, preds, uids))
    metrics.update(spearman_correlation_scorer(golds, probs, preds, uids))
    metrics["pearson_spearman"] = np.mean(
        [metrics["pearson_correlation"], metrics["spearman_correlation"]]
    )
    return metrics



import numpy as np

from emmental.metrics.accuracy import accuracy_scorer
from emmental.metrics.fbeta import f1_scorer


def accuracy_f1_scorer(golds, probs, preds, uids=None):
    """Average of accuracy and f1 score.

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities. (Not used!)
    :type probs: k-d np.array
    :param preds: Predicted target values.
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :return: Average of accuracy and f1.
    :rtype: dict
    """

    metrics = dict()
    metrics.update(accuracy_scorer(golds, probs, preds, uids))
    metrics.update(f1_scorer(golds, probs, preds, uids))
    metrics["accuracy_f1"] = np.mean([metrics["accuracy"], metrics["f1"]])

    return metrics



import numpy as np


def accuracy_scorer(golds, probs, preds, uids=None, normalize=True):
    """Accuracy classification score.

    :param golds: Ground truth (correct) target values.
    :type golds: 1-d np.array
    :param probs: Predicted target probabilities. (Not used!)
    :type probs: k-d np.array
    :param preds: Predicted target values.
    :type preds: 1-d np.array
    :param uids: Unique ids.
    :type uids: list
    :param normalize: Normalize the results or not, defaults to True
    :param normalize: bool, optional
    :return: Accuracy, if normalize == True, return the fraction of correctly
        predicted samples (float), else returns the number of correctly predicted
        samples (int).
    :rtype: dict
    """

    if normalize:
        return {"accuracy": np.where(golds == preds)[0].shape[0] / golds.shape[0]}
    else:
        return {"accuracy": np.where(golds == preds)[0].shape[0]}



import torch.nn as nn


class IdentityModule(nn.Module):
    """An identity module that outputs the input."""

    def __init__(self):
        super().__init__()

    def forward(self, x):
        """Forward function.
        :param x: Input tensor.
        :type x: torch.Tensor
        :return: Output of identity module which is the same with input.
        :rtype: torch.Tensor
        """

        return x



import json

from tensorboardX import SummaryWriter

from emmental.logging.log_writer import LogWriter
from emmental.meta import Meta


class TensorBoardWriter(LogWriter):
    """A class for logging to Tensorboard during training process."""

    def __init__(self):
        super().__init__()

        # Set up tensorboard summary writer
        self.writer = SummaryWriter(Meta.log_path)

    def add_scalar(self, name, value, step):
        """Log a scalar variable.

        :param name: The name of the scalar
        :type name: str
        :param value: The value of the scalar
        :type value: float or int
        :param step: The current step
        :type step: float or int
        """

        self.writer.add_scalar(name, value, step)

    def write_config(self, config_filename="config.yaml"):
        """Dump the config to file.

        :param config_filename: The config filename, defaults to "config.yaml"
        :type config_filename: str, optional
        """

        config = json.dumps(Meta.config)
        self.writer.add_text(tag="config", text_string=config)

        super().write_config(config_filename)

    def close(self):
        """Close the tensorboard writer.
        """

        self.writer.close()
